/* This implementation uses a singly-linked list data-structure. Notice
 * that the underlying implementation is completely hidden from the client, and 
 * it may be changed to, say, an array-based queue without affecting client code. */
#include "queue.h"
#include <stdlib.h>

typedef struct QueueNode QueueNode;

struct Queue {
    QueueNode *head;
    QueueNode *front;
    QueueNode *back;
};

struct QueueNode
{
    QueueType item;
    QueueNode *next;
};


Queue *queue_create(int max)
/* Creates a queue with capacity for max items. 
 * Returns pointer to queue object if successful, NULL if fails. */
{
    Queue *q = (Queue *)malloc(sizeof(Queue));
    if (q) {
        /* Allocate link-list nodes in one go (as an array). This is more efficient
         * than allocating separately, and makes deallocation simpler. */
        q->head = (QueueNode *)malloc(sizeof(QueueNode) * (max+1)); /* notice +1 */
        if (!q->head) {
            free(q);
            return NULL;
        }

        /* Connect array elements into linked list */
        q->back = q->head + max;
        for (q->front = q->head; q->front != q->back; ++q->front)
            q->front->next = q->front + 1;
        q->back->next = q->head; /* make list circular */
        /* Note: front == back, so queue is initialised as empty */
    }
    return q;
    
    /* Implementation note: This queue is marked as empty when the back == front. The
     * queue is marked as full when the front is one node *behind* the back. For this 
     * reason we allocate max+1 nodes so that the queue holds max items when full. The
     * additional node acts as a placeholder, so we don't need to maintain a "number-of-
     * elements" counter to determine when the queue is full or empty. */        
}

void queue_destroy(Queue *q)
/* Releases all memory associated with the queue. */
{
    free(q->head);
    free(q);
}

int queue_push(Queue *q, QueueType item)
/* Push item onto front of queue. Returns 0 if successful, -1 if queue is full. */
{
    if (q->front->next != q->back) { /* full when front->next == back */
        q->front->item = item;
        q->front = q->front->next;
        return 0;
    }
    return -1;
}

int queue_pop(Queue *q, QueueType *item)
/* Pop item off back of queue. Returns 0 if successful, -1 if queue is empty. */
{
    if (q->back != q->front) { /* empty when back == front */
        *item = q->back->item;
        q->back = q->back->next;
        return 0;
    }
    return -1;
}
